/* This file was generated by syncqt. */
#ifndef QT_QTPACKETPROTOCOL_VERSION_H
#define QT_QTPACKETPROTOCOL_VERSION_H

#define QTPACKETPROTOCOL_VERSION_STR "5.15.18"

#define QTPACKETPROTOCOL_VERSION 0x050F12

#endif // QT_QTPACKETPROTOCOL_VERSION_H
